from bs4 import BeautifulSoup as bs
import requests

from typing import Optional, List, Dict, Union

class Caption():
    """Caption object of video"""
    def __init__(self, language_code: str, name: str, url, translatable: str = False, translate_langs: Optional[List] = None):
        self.language_code = language_code
        self.name = name
        self._url = url
        self.is_translatable = translatable
        self._trans_lang = translate_langs
    
    def __repr__(self):
        return f'<Caption lang="{self.name}" code="{self.language_code}" is_translatable={self.is_translatable}>'

    @property
    def available_translations(self) -> 'TransLangQuery':
        """Return all available translations of the caption
        
        :return: List of all available translations
        :rtype: List[TransLangQuery]
        """
        if self.is_translatable:
            return TransLangQuery(TranslationLang(i) for i in self._trans_lang)

    @property
    def xml(self) -> str:
        """Raw XML format of the caption

        :return: Caption in XML format
        :rtype: str
        """
        return requests.get(self._url).text

    def get_text(self, delimiter='\n') -> str:
        """Extract text from XML string

        :param delimiter: Delimiter for spliting sentences, defaults to '\\n'
        :type delimiter: str, optional
        :return: Caption in pure text form
        :rtype: str
        """
        raw = bs(self.xml, 'lxml')
        text = raw.findAll('text')
        string = [bs(i.text, 'lxml').text for i in text]
        return delimiter.join(string)

    @property
    def dict(self) -> Dict[str, Union[str, float]]:
        """GIves you a dictionary containing the captions content and metadata

        :return: Caption text and it's metadata in dictionary format
        :rtype: dict
        """
        raw = bs(self.xml, 'lxml')
        text = raw.findAll('text')
        return [{
            'start_from': float(i['start']),
            'duration': float(i['dur']),
            'text': bs(i.text, 'lxml').text
        } for i in text]

    def translate_to(self, language_code: str) -> Optional['TranslatedCaption']:
        """Translate the caption to the given language if the caption is translatable

        :param language_code: Language code of targeted language for translation
        :type language_code: str
        :return: Translated caption
        :rtype: TranslatedCaption
        """
        if self.is_translatable:
            if not language_code in self.available_translations.get_language_code():
                return None
            return TranslatedCaption(self.available_translations.get_language(language_code), self._url, self.language_code)

class TranslatedCaption(Caption):
    """Object of translated version of caption"""
    def __init__(self, language: 'TranslateLang', url: str, original_lang_code: str):
        super(TranslatedCaption, self).__init__(language.language_code, language.name , url)
        del Caption.translate_to

        self.language_code = language.language_code
        self.name = language.name
        self._url = url+'&tlang='+self.language_code
        self.original_language_code = original_lang_code
    
    def __repr__(self):
        return f'<TranslatedCaption lang="{self.name}" code="{self.language_code}" translated_from="{self.original_language_code}">'

class TranslationLang:
    """Translation language for translatable caption"""
    def __init__(self, raw: dict):
        self.name = raw["languageName"]["simpleText"]
        self.language_code = raw["languageCode"]

    def __repr__(self):
        return f'<TranslationLang name="{self.name}" code="{self.language_code}">'

class CaptionQuery(list):
    """Container of available captions of the video"""
    def __init__(self, data: list, default: int = 0):
        super(CaptionQuery, self).__init__(data)
        self.default=default

    def __repr__(self):
        return f'<CaptionQuery {list(self)}>'

    def get_caption(self, language_code: Optional[str] = None) -> Optional[Caption]:
        """Get caption by language code

        :param language_code: Language code of the caption, defaults set to None
        :type language_code: str, optional
        :return: Caption that corresponds to the language code. Return caption with default language if language_code is not specified. Return None if caption with the language code is not found.
        :rtype: Caption or None
        """
        if not language_code:
            return self[self.default]
        else:
            for i in self:
                if i.language_code == language_code:
                    return i

class TransLangQuery(list):
    """Container of language available for translation of a caption
    """
    def __init__(self, data: list):
        super(TransLangQuery, self).__init__(data)

    def get_language(self, language_code: str) -> Optional[TranslationLang]:
        """Get language by language code

        :param language_code: Language code of the language
        :type language_code: str
        :return: Language that corresponds to the language code. Return None if caption with the language code is not found.
        :rtype: List[TranslateLang] or None
        """
        for i in self:
            if i.language_code == language_code:
                return i

    def get_name(self) -> list:
        """Return name of all available languages
        
        :return: Name of all available languages
        :rtype: list
        """
        return [i.name for i in self]

    def get_language_code(self) -> list:
        """Return language code of all available languages

        :return: Language code of all available languages
        :rtype: list
        """
        return [i.language_code for i in self]


BASE_URL = "https://www.youtube.com"
SEARCH_BASE_URL = "https://www.youtube.com/results?search_query="
SEARCH_CONTINUATION_URL = "https://www.youtube.com/youtubei/v1/search?key="
PLAYLIST_BASE_URL = "https://www.youtube.com/playlist?list="
PLAYLIST_CONTINUTION_URL = "https://www.youtube.com/youtubei/v1/browse?key="
CHANNEL_USERNAME_URL = "https://www.youtube.com/user/"
CHANNEL_ID_URL = "https://www.youtube.com/channel/"
VIDEO_PLAYER_URL = "https://www.youtube.com/watch?v="
COMMENT_AJAX_URL = 'https://www.youtube.com/comment_service_ajax'
QUERY_SUGGESTIONS_URL = 'https://suggestqueries-clients6.youtube.com/complete/search'


from youtube_scraping_api.utils import search_dict, get_thumbnail
from youtube_scraping_api.parser.video import Video
from youtube_scraping_api.decorators import custom_property

def cleanupData(data, nextCT=None):
    result = []
    for i in data:
        try: typeOfRenderer = list(i.keys())[0]
        except: raise
        if typeOfRenderer == "continuationItemRenderer":
            continue
        each = i[typeOfRenderer]
        eachFinal = RENDERER_PARSER[typeOfRenderer](each)
        result.append(eachFinal)
    if len(result) == 1: return result[0]
    return result

class Playlist(list):
    """A container of playlist metadata and  its videos"""
    def __init__(self, response, builtin_called=False):
        self.first_data = response["metadata"]["playlistMetadataRenderer"]
        self.second_data = next(search_dict(response, "videoOwnerRenderer"))
        self._is_builtin_called = builtin_called
        self._static_properties = []
        self._has_generated = False

        data = cleanupData(next(search_dict(response,"itemSectionRenderer"))["contents"])
        super(Playlist, self).__init__(data)

    def parse_data(self):
        pass

    @custom_property
    def title(self):
        """Exrtract name of the playlist
        
        :return: Playlist name
        :rtype: str
        """
        return self.first_data["title"]

    @custom_property
    def description(self):
        """Extract description of the playlist if available
    
        :return: Playlist description
        :rtype: str of None
        """
        return self.first_data["description"] if "description" in self.first_data else None
    
    @custom_property
    def owner(self):
        """Return the name of playlist creator
    
        :return: Playlist creator name
        :rtype: str
        """
        return self.second_data["title"]["runs"][0]["text"]
    
    @custom_property
    def video_count(self):
        """Count how many videos are in the playlist

        :return: Number of videos in the playlist
        :rtype: int
        """
        video_count, *_ = next(search_dict(self.response, "stats"))
        return int(video_count["runs"][0]["text"].replace(",", ""))
    
    @custom_property
    def view_count(self):
        """Count the total views of all videos in the playlist

        :return: Total views of videos in the playlist
        :rtype: int
        """
        _, total_views, _ = next(search_dict(self.response, "stats"))
        return int(total_views["simpleText"].split()[0].replace(",", ""))
    
    @custom_property
    def last_updated(self):
        """Return the time when the playlist is last updated

        :return: Playlist last updated time
        :rtype: str
        """
        *_, last_updated = next(search_dict(self.response, "stats"))
        last_updated = "".join(i["text"] for i in last_updated["runs"])
        if "Last updated on" in last_updated: last_updated = " ".join(last_updated.split()[3:])
        if "Updated" in last_updated: last_updated = " ".join(last_updated.split()[1:])
        return last_updated

    def __repr__(self):
        return f'<Playlist title="{self.title}" video_count={len(self)}>'

class PlaylistVideo(Video):
    """A container of playlist with the video index in playlist video that function exactly the same as Video Object"""
    def __init__(self, data):
        self.index = int(data["index"]["simpleText"])
        """Index of the video in playlist

        :return: Video index in playlist
        :rtype: int
        """
        super().__init__(
            data["videoId"],
            length = data["lengthText"]["simpleText"] if "lengthText" in data else None,
            thumbnails = get_thumbnail(data["videoId"]),
            builtin_called = True
        )
    
    def __repr__(self):
        return f'<PlaylistVideo index={self.index} id={self.id}>'

    @property
    def raw(self):
        """Return a dictionary containing all data of the playlist video

        :return: Raw data of video
        :rtype: dict
        """
        return {"index": self.index, **super().raw}

RENDERER_PARSER = {
    "playlistVideoListRenderer": lambda data: cleanupData(data["contents"]),
    "playlistVideoRenderer": PlaylistVideo
}


import json
import string as strlib
import requests
from bs4 import BeautifulSoup as bs
from youtube_scraping_api.constants import HEADERS, THUMBNAIL_TEMPLATE

def search_dict(partial, key):
    """Recursive search in dictionary

    :param partial: Dictionary to search in
    :type partial: dict
    :param key: Key that you want to search in dictionary
    :type key: str
    :return: Value in dictionary of targeted key
    :rtype: Any
    """
    if isinstance(partial, dict):
        for k, v in partial.items():
            if k == key:
                yield v
            else:
                for o in search_dict(v, key):
                    yield o
    elif isinstance(partial, list):
        for i in partial:
            for o in search_dict(i, key):
                yield o

def find_snippet(text, start, end, skip=(0, 0)):
    """Find snippet in text

    :param text: Text to search in
    :type text: str
    :param start: Where to start grabbing text
    :type start: str
    :param end: Where to stop grabbing text and return
    :type end: str
    :param skip: Number of character to trim in front and behind gragbbed text
    :type skip: tuple
    :return: Snippet found in the text
    :rtype: str
    """
    start_index = text.find(start)
    if start_index == -1:
        return start_index
    end = text.find(end, start_index)
    return text[start_index + len(start) + skip[0]:end - skip[1]]

def parse_continuation_token(data):
    """Extract continuation from raw JSON data

    :param data: Raw JSON data
    :type data: dict
    :return: Continuation token
    :rtype: str
    """
    try: nextCT = next(search_dict(data, "token"))
    except: nextCT = None
    finally: return nextCT

def convert_valid_filename(string):
    """Remove invalid character for saving file from string

    :param string: String to be converted into valid filename
    :type string: str
    :return: String that has invalid character removed
    :rtype: str
    """
    valid_chars = "-_.() %s%s" % (strlib.ascii_letters, strlib.digits)
    return "".join(c for c in string if c in valid_chars)

def get_initial_data(html):
    """Extract primary JSON data from raw HTML source code

    :param html: Raw HTML source code
    :type html: str
    :return: JSON data in form of dictionary
    :rtype: dict
    """
    return json.loads(find_snippet(html, "var ytInitialData = ", "</script>", (0, 1)))

def get_initial_player_response(html):
    """Extract JSON data where video download links are located

    :param html: Raw HTML source code
    :type html: str
    :return: JSON data in form of dictionary
    :rtype: dict
    """
    return json.loads(find_snippet(html, "var ytInitialPlayerResponse = ", ";</script>", (0, 1))+"}", strict=False)

def reveal_redirect_url(url):
    """Get real url from redirect url

    :param url: Redirect url
    :type url: str
    :return: Real url
    :rtype: str
    """
    return bs(requests.get(url, headers=HEADERS).content, "lxml").find("div", {"id": "redirect-action-container"}).find("a")["href"]

def get_thumbnail(videoId):
    """Get url for thumbnails of video

    :param videoId: Youtube ID of the video
    :type videoId: str
    :return: A dictionary of thumbnail urls
    :rtype: dict
    :todo: Check thumbnail urls availability
    """
    return dict(map(lambda i: (i[0], i[1].format(videoId)), THUMBNAIL_TEMPLATE.items()))

def get_proxy():
    proxies = requests.get('https://api.proxyscrape.com/v2/?request=getproxies&protocol=http&timeout=4550&country=all&ssl=all&anonymity=all&simplified=true').text.split('\r\n')
    for i in proxies:
        try:
            proxy = {
                'http': 'http://' + i,
                'https': 'http://' + i
            }
            requests.get('https://youtube.com', proxies=proxy, timeout=5)
            return proxy
        except:
            pass



import re
from itertools import chain

class Cipher:
    def __init__(self, js):
        self.transform_plan = self.get_transform_plan(js)
        var, _ = self.transform_plan[0].split(".")
        self.transform_map = self.get_transform_map(js, var)
        self.js_func_regex = re.compile(r"\w+\.(\w+)\(\w,(\d+)\)")

    def get_signature(self, ciphered_signature):
        signature = list(ciphered_signature)

        for js_func in self.transform_plan:
            name, argument = self.parse_function(js_func)  # type: ignore
            signature = self.transform_map[name](signature, argument)
        return "".join(signature)

    def parse_function(self, js_func: str):
        parse_match = self.js_func_regex.search(js_func)
        if not parse_match:
            raise Exception
        fn_name, fn_arg = parse_match.groups()
        return fn_name, int(fn_arg)

    def regex_search(self, pattern, string, group):
        regex = re.compile(pattern)
        results = regex.search(string)
        if not results:
            raise Exception
        return results.group(group)

    def get_initial_function_name(self, js):
        function_patterns = [
            r"\b[cs]\s*&&\s*[adf]\.set\([^,]+\s*,\s*encodeURIComponent\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
            r"\b[a-zA-Z0-9]+\s*&&\s*[a-zA-Z0-9]+\.set\([^,]+\s*,\s*encodeURIComponent\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
            r'(?:\b|[^a-zA-Z0-9$])(?P<sig>[a-zA-Z0-9$]{2})\s*=\s*function\(\s*a\s*\)\s*{\s*a\s*=\s*a\.split\(\s*""\s*\)',  # noqa: E501
            r'(?P<sig>[a-zA-Z0-9$]+)\s*=\s*function\(\s*a\s*\)\s*{\s*a\s*=\s*a\.split\(\s*""\s*\)',  # noqa: E501
            r'(["\'])signature\1\s*,\s*(?P<sig>[a-zA-Z0-9$]+)\(',
            r"\.sig\|\|(?P<sig>[a-zA-Z0-9$]+)\(",
            r"yt\.akamaized\.net/\)\s*\|\|\s*.*?\s*[cs]\s*&&\s*[adf]\.set\([^,]+\s*,\s*(?:encodeURIComponent\s*\()?\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
            r"\b[cs]\s*&&\s*[adf]\.set\([^,]+\s*,\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
            r"\b[a-zA-Z0-9]+\s*&&\s*[a-zA-Z0-9]+\.set\([^,]+\s*,\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
            r"\bc\s*&&\s*a\.set\([^,]+\s*,\s*\([^)]*\)\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
            r"\bc\s*&&\s*[a-zA-Z0-9]+\.set\([^,]+\s*,\s*\([^)]*\)\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
            r"\bc\s*&&\s*[a-zA-Z0-9]+\.set\([^,]+\s*,\s*\([^)]*\)\s*\(\s*(?P<sig>[a-zA-Z0-9$]+)\(",  # noqa: E501
        ]
        for pattern in function_patterns:
            regex = re.compile(pattern)
            function_match = regex.search(js)
            if function_match:
                return function_match.group(1)
        raise Exception

    def get_transform_plan(self, js):
        name = re.escape(self.get_initial_function_name(js))
        pattern = r"%s=function\(\w\){[a-z=\.\(\"\)]*;(.*);(?:.+)}" % name
        return self.regex_search(pattern, js, group=1).split(";")

    def get_transform_object(self, js, var):
        pattern = r"var %s={(.*?)};" % re.escape(var)
        regex = re.compile(pattern, flags=re.DOTALL)
        transform_match = regex.search(js)
        if not transform_match:
            raise Exception
        return transform_match.group(1).replace("\n", " ").split(", ")

    def get_transform_map(self, js, var):
        transform_object = self.get_transform_object(js, var)
        mapper = {}
        for obj in transform_object:
            # AJ:function(a){a.reverse()} => AJ, function(a){a.reverse()}
            name, function = obj.split(":", 1)
            fn = self.map_functions(function)
            mapper[name] = fn
        return mapper

    def reverse(self, arr, _):
        return arr[::-1]

    def splice(self, arr, b):
        return arr[b:]

    def swap(self, arr, b):
        r = b % len(arr)
        return list(chain([arr[r]], arr[1:r], [arr[0]], arr[r + 1 :]))

    def map_functions(self, js_func):
        mapper = (
            # function(a){a.reverse()}
            (r"{\w\.reverse\(\)}", self.reverse),
            # function(a,b){a.splice(0,b)}
            (r"{\w\.splice\(0,\w\)}", self.splice),
            # function(a,b){var c=a[0];a[0]=a[b%a.length];a[b]=c}
            (r"{var\s\w=\w\[0\];\w\[0\]=\w\[\w\%\w.length\];\w\[\w\]=\w}", self.swap),
            # function(a,b){var c=a[0];a[0]=a[b%a.length];a[b%a.length]=c}
            (
                r"{var\s\w=\w\[0\];\w\[0\]=\w\[\w\%\w.length\];\w\[\w\%\w.length\]=\w}",
                self.swap,
            ),
        )

        for pattern, fn in mapper:
            if re.search(pattern, js_func):
                return fn

        raise Exception



from youtube_scraping_api.urls import *
from youtube_scraping_api.utils import *
from youtube_scraping_api.decorators import *
from youtube_scraping_api.cipher import Cipher
from youtube_scraping_api.constants import HEADERS
from youtube_scraping_api.parser.channel import Channel
from youtube_scraping_api.caption import CaptionQuery, Caption

import os
import time
import requests
from tqdm import tqdm
from urllib.parse import parse_qs
from typing import List, Dict, Any, Optional, Callable, Generator

class Video():
    """Container for video data"""
    def __init__(self, videoId: str, builtin_called: bool = False, **kwargs):
        self._session = requests.Session()
        self._session.headers = HEADERS
        self._has_generated = False

        self._raw = None
        self._player_data = None
        self._init_data = None
        self._primary_info = None
        self._secondary_info = None
        self._player_info = None
        self._is_builtin_called = builtin_called

        self.id = videoId
        self.thumbnails = get_thumbnail(self.id)

        self._static_properties = kwargs

    def __repr__(self):
        return f'<Video id="{self.id}" title="{self.title}" author="{self.author.name}">'

    def parse_data(self) -> None:
        """Fetch HTML source code and extract JSON data from it

        :return: Nothing, data have been set inside local variable
        :rtype: None
        """
        try:
            self._raw = self._session.get(VIDEO_PLAYER_URL+self.id).text
            self._player_data = get_initial_player_response(self._raw)
            self._init_data = get_initial_data(self._raw)
            self._primary_info = next(search_dict(self._init_data, "videoPrimaryInfoRenderer"))
            self._secondary_info = next(search_dict(self._init_data, "videoSecondaryInfoRenderer"))
            self._player_info = self._player_data["videoDetails"]

            self._has_generated = True

        except:
            self._session = requests.Session()
            self._session.headers = HEADERS
            time.sleep(3)
            self.parse_data()

    def get_signature_url(self, url: str) -> str:
        """Get decrypted download link for the video

        :param url: Encrypted download link of the video
        :type url: str
        :return: Usable download link of video
        :rtype: str
        :note: This function isn't developed by me since I have no enough time to dive so deep into Javascript. Credit to PyTube.
        """
        base_url = find_snippet(self._raw, "jsUrl", ",", (3, 1))
        js_url = BASE_URL+base_url
        js_content = self._session.get(js_url).text
        cipher = Cipher(js_content)
        s, sp, url = [i[0] for i in parse_qs(url).values()]
        return url + "&sig=" + cipher.get_signature(s)

    def get_comment_count(self) -> int:
        """Fetch the amount of comments of the video

        :return: Number of comments
        :rtype: int
        """
        continuation = next(search_dict(self._init_data, 'nextContinuationData'))['continuation'].replace('%3D', '=')
        xsrf_token = find_snippet(self._raw, 'XSRF_TOKEN', ",", (3, 1)).replace(r'\u003d', '\u003d')
        params = {
            'action_get_comments': 1,
            'pbj': 1,
            'ctoken': continuation,
            'continuation': continuation,
            'type': 'next'
        }
        data = self._session.post(COMMENT_AJAX_URL, data={'session_token': xsrf_token}, params=params).json()
        comment_count = int(next(search_dict(data, 'countText'))['runs'][0]['text'].replace(',', ''))
        return comment_count

    @custom_property
    def title(self) -> str:
        """Extract video title

        :return: Video title
        :rtype: str
        """
        return "".join(i["text"] for i in self._primary_info["title"]["runs"])

    @custom_property
    def type(self) -> str:
        """Determine type of video

        :return: Video type
        :rtype: str
        """
        return "livestream" if "isLive" in self._primary_info["viewCount"]["videoViewCountRenderer"] and self._primary_info["viewCount"]["videoViewCountRenderer"]["isLive"] else "video"

    @custom_property
    def supertitle(self) -> Optional[List[Dict[str, str]]]:
        """Extract supertitle(custom tags) from video

        :return: List of supertitles of video
        :rtype: list
        """
        if not "superTitleLink" in self._primary_info: return None
        supertitle = [{
            'text': i['text'].strip(),
            'url': next(search_dict(i, 'url'))
        } for i in self._primary_info["superTitleLink"]["runs"] if i['text'].strip()]
        return supertitle

    @custom_property
    def description(self) -> Optional[str]:
        """Extract full description of video

        :return: Description of video
        :rtype: str
        """
        return "".join(i["text"] for i in self._secondary_info["description"]["runs"]) if "description" in self._secondary_info else None

    @custom_property
    def tags(self) -> List[str]:
        """Extract descriptive keywords which content creators can add to thier video to help viewers find their content

        :return: List of all tags of video
        :rtype: list
        """
        return self._player_info["keywords"] if "keywords" in self._player_info else None

    @custom_property
    def publish_time(self) -> str:
        """Extract the time when the video is published

        :return: Time when video is published
        :rtype: str
        :todo: Convert output string to datetime object
        """
        return self._primary_info["dateText"]["simpleText"]

    @custom_property
    def author(self) -> Channel:
        """Extract the content creator who upload the video

        :return: Video author
        :rtype: Channel
        """
        return Channel(
            name = self._player_info["author"],
            url = self._secondary_info["owner"]["videoOwnerRenderer"]["navigationEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"],
            channel_id = self._player_info["channelId"]
        )

    @custom_property
    def length(self) -> int:
        """Extract the length of the video in second

        :return: Length of video
        :rtype: int
        """
        return int(self._player_info["lengthSeconds"])

    @custom_property
    def view_count(self) -> Optional[int]:
        return self._player_info["viewCount"]

    @custom_property
    def raw(self) -> Dict[str, Any]:
        """Return a dictionary containing all data of video

        :return: Raw data of video
        :rtype: dict
        """
        primary_info = self._primary_info
        player_info = self._player_info
        cleaned_data = {
            "videoId": self.id,
            "type": self.type,
            "title": self.title,
            "supertitle": self.supertitle,
            "description": self.description,
            "tags": self.tags,
            "publish_time": self.publish_time,
            "author": self.author.name,
            "length": self.length,
            "thumbnails": self.thumbnails,
            "statistics": {
                "view_count": int(player_info["viewCount"]),
                **dict(zip(["like_count", "unlike_count"], map(lambda i: int(i.replace(",", "")), primary_info["sentimentBar"]["sentimentBarRenderer"]["tooltip"].split(" / ")))),
                "comment_count": self.get_comment_count()
            }
        }
        return cleaned_data

    @custom_property
    def download_data(self) -> Optional[Dict[int, Dict[str, Any]]]:
        """Parse download links and metadata of the video

        :return: List of dictionary containing download links and metadata
        :rtype: list
        """
        if not self._player_data: self.parse_data()
        try:
            return dict([(i["itag"], {
                "url": i["url"] if "url" in i else None,
                'signature_cipher': i["signatureCipher"] if "signatureCipher" in i else None,
                "mime_type": i["mimeType"],
                "bitrate": i["bitrate"],
                "width": i["width"] if "width" in i else None,
                "height": i["height"] if "height" in i else None,
                "size": i["contentLength"] if "contentLength" in i else None,
                "fps": i["fps"] if "fps" in i else None,
                "quality": i["quality"],
                "quality_label": i["qualityLabel"] if "qualityLabel" in i else None,
                "duration": i["approxDurationMs"] if "approxDurationMs" in i else None
            }) for i in self._player_data["streamingData"]["formats"]+self._player_data["streamingData"]["adaptiveFormats"]]) if self.type!="livestream" else None
        except:
            pass

    def get_file_size(self, url: str) -> int:
        """Get the size of video stream

        :param url: Download link of the video
        :type url: str
        :return: size of video stream in bytes
        :rtype: int
        """
        return int(self._session.get(url, stream=True).headers.get('content-length', 0))

    def stream(self, url, chunk_size: int = 8192, range_size: int = 10000000000) -> Generator:
        """Request and yield chunks of video stream

        :param url: Download link of video stream
        :type url: str
        :param chunk_size: Size of chunk per request
        :type chunk_size: int
        :param range_size: Default size to download, can be overridden
        :type range_size: int, optional
        :return: video stream chunk
        :rtype: bytes
        """
        file_size: int = self.get_file_size(url)

        downloaded = 0
        while downloaded < file_size:
            stop_pos = min(downloaded + range_size, file_size) - 1
            range_header = f"bytes={downloaded}-{stop_pos}"
            response = requests.get(url, headers={"Range": range_header}, stream=True)
            for chunk in response.iter_content(chunk_size):
                if not chunk: break
                downloaded += len(chunk)
                yield chunk
        return

    def download(self, itag: int = None, path: str = ".", log_progress: bool = True, chunk_size: int = 4096, callback_func: Optional[Callable[[Any], None]] = None, name: Optional[str] = None) -> None:
        """Download video from YouTube into local storage

        :param itag: Itag of the video to download, video with best quality will be downloaded if set to None, default set to None
        :type itag: int, optional
        :param path: Relative or absolute path to save the video
        :type path: str, optional
        :param log_progress: Wether to show progress bar of download or not. Default set to True
        :type log_progress: bool, optional
        :param chunk_size: Size of chunk per request. Default set to 4096
        :type chunk_size: int, optional
        :param callback_func: Callback function to be called downloading video. Default set to None
        :type callback_func: Callable, optional
        :param name: Filename of video. Use video title if not set
        :type name: str, optional
        :return: None, just download the video and save it
        :rtype: None
        """
        if not self.download_data:
            return None

        if itag:
            if itag in self.download_data.keys():
                target = self.download_data[itag]
            else:
                raise RuntimeError('Itag does not exist!')
        else:
            target = list(self.download_data.values())[0]

        if target['url'] != None:
            target_url = target['url']
        elif target["signature_cipher"]:
            target_url = self.get_signature_url(target["signature_cipher"])

        vid_name = name if name else convert_valid_filename(self.title)
        extension = target["mime_type"].split(";")[0].split("/")[-1]


        if log_progress:
            print(self.title)
            progress_bar = tqdm(total=self.get_file_size(target_url), unit='iB', unit_scale=True)

        with open(os.path.join(path, f"{vid_name}.{extension}"), "wb") as f:
            for chunk in self.stream(target_url, chunk_size=chunk_size):
                if log_progress: progress_bar.update(len(chunk))
                f.write(chunk)

    @custom_property
    def captions(self) -> CaptionQuery:
        """Give you a list of available captions for the video

        :return: List of available captions
        :rtype: CaptionQuery
        """
        if not self._player_data:
            self.parse_data()

        if not "captions" in self._player_data:
            return None
        raw = self._player_data["captions"]["playerCaptionsTracklistRenderer"]
        default_raw = raw["audioTracks"][0]
        default = default_raw["defaultCaptionTrackIndex"] if "defaultCaptionTrackIndex" in default_raw else 0
        caption_list = raw["captionTracks"]
        result = CaptionQuery((Caption(i["languageCode"], i["name"]["simpleText"], i["baseUrl"], i["isTranslatable"], raw["translationLanguages"]) for i in caption_list), default)

        return result



import json

import requests
from .utils import get_initial_data, search_dict
from .urls import BASE_URL
from .constants import HEADERS

class AvailableSearchFilter:
	"""An object containing all available search filters
	"""
	def __init__(self, duration=None, upload_date=None, type=None, features=None, sort_by=None):
		self.duration = duration
		self.upload_date = upload_date
		self.type = type
		self.features = features
		self.sort_by = sort_by

	def __repr__(self):
		return f'<AvailableSearchFilter sort_by:{len(self.sort_by)} upload_date:{len(self.upload_date)} type:{len(self.type)} features:{len(self.features)} duration:{len(self.duration)}>'

class SearchFilter:
	"""Filter for search results

	:param type:
		(optional) Type of search result
	:type type: str or None
	:param features:
		(optional) Features of vidoes
	:type features: list or None
	:param sort_by:
		(optional) Criteria of sorting search results
	:type sort_by: str or None
	:param upload_date:
		(optional) Upload date of videos
	:type upload_date: str or None
	:param duration:
		(optional) Expected duration of videos
	:type upload_date: str or None

	:rtype: Object[SearchFilter]
	"""
	def __init__(self, type=None, features=None, sort_by=None, upload_date=None, duration=None):
		self.type = (type, 'Type')
		self.features = (features, 'Features')
		self.sort_by = (sort_by, 'Sort by')
		self.upload_date = (upload_date, 'Upload date')
		self.duration = (duration, 'Duration')

	@classmethod
	def get_all_filters(self):
		"""Get all available filters that can be used when querying search results
		"""
		session = requests.Session()
		session.headers = HEADERS
		raw = get_initial_data(session.get('https://www.youtube.com/results?search_query=hermitcraft').text)
		filter_groups = [i['searchFilterGroupRenderer'] for i in next(search_dict(raw, "searchSubMenuRenderer"))['groups']]
		cleaned_filter_groups = dict([[i['title']['simpleText'].lower().replace(' ', '_'), [i['searchFilterRenderer']['label']['simpleText'] for i in i['filters']]] for i in filter_groups])

		return AvailableSearchFilter(**cleaned_filter_groups)

def get_filtered_url(session, base_url, filter):
	"""Generate valid search url that includes query string filter

	:param session: Requests session
	:type session: Session
	:param base_url: Base search url that includes only query string
	:type base_url: str
	:param filter: Search filter that defined by user
	:type filter: SearchFilter
	:return: Valid search url
	:rtype: str
	"""
	url = base_url
	for i in [filter.type, filter.sort_by, filter.upload_date, filter.duration]:
		if i[0]:
			raw = get_initial_data(session.get(url).text)
			filter_groups = [i['searchFilterGroupRenderer'] for i in next(search_dict(raw, "searchSubMenuRenderer"))['groups']]
			target_group = [f for f in filter_groups if f['title']['simpleText']==i[1]][0]
			if not target_group: raise AttributeError(f'Filter type {i} not found')
			filters = [i['searchFilterRenderer'] for i in target_group['filters']]
			target_filter = [f for f in filters if f['label']['simpleText']==i[0]]
			if not target_filter: raise AttributeError('Filter "{}" not found in {}'.format(i[0], target_group['title']['simpleText']))
			try: url = BASE_URL+next(search_dict(target_filter, 'url'))
			except: pass

	if filter.features[0] and isinstance(filter.features[0], list):
		for i in filter.features[0]:
			if i and isinstance(i, str):
				raw = get_initial_data(session.get(url).text)
				filter_groups = [i['searchFilterGroupRenderer'] for i in next(search_dict(raw, "searchSubMenuRenderer"))['groups']]
				target_group = [f for f in filter_groups if f['title']['simpleText']=='Features'][0]
				if not target_group: raise AttributeError(f'Filter type {i} not found')
				filters = [i['searchFilterRenderer'] for i in target_group['filters']]
				target_filter = [f for f in filters if f['label']['simpleText']==i]
				if not target_filter: raise AttributeError('Filter "{}" not found in {}'.format(i[0], target_group['title']['simpleText']))
				try: url = BASE_URL+next(search_dict(target_filter, 'url'))
				except: pass
			else:
				raise TypeError('Features filter elements must be a string')
	return url



PAYLOAD = {
    "context": {
        "client": {
            "clientName": "WEB",
            "clientVersion": "2.20201220.08.00",
        },
        "user": {
            "lockedSafetyMode": False,
        }
    },
    "webClientInfo": {
        "isDocumentHidden": True
    }
}

HEADERS = {
    "x-youtube-client-name": "1",
    "x-youtube-client-version": "2.20210407.08.00",
    "accept-language": "en-US"
}

THUMBNAIL_TEMPLATE = {
    "default": "https://i.ytimg.com/vi/{}/default.jpg",
    "medium": "https://i.ytimg.com/vi/{}/mqdefault.jpg",
    "high": "https://i.ytimg.com/vi/{}/hqdefault.jpg",
    "standard": "https://i.ytimg.com/vi/{}/sddefault.jpg",
    "maxres": "https://i.ytimg.com/vi/{}/maxresdefault.jpg"
}



from ..utils import get_thumbnail, search_dict
from collections import Counter
from .video import Video
from .channel import Channel
from ..urls import BASE_URL
import requests

def cleanupData(data, nextCT=None, to_object=False):
    result = []
    for i in data:
        try: typeOfRenderer = list(i.keys())[0]
        except: raise
        each = i[typeOfRenderer]
        try: typeOfRenderer = "liveStreamRenderer" if each["badges"][0]["metadataBadgeRenderer"]["label"] == "LIVE NOW" else typeOfRenderer
        except: pass
        try: typeOfRenderer = "premiereRenderer" if each["badges"][0]["metadataBadgeRenderer"]["label"] == "PREMIERING NOW" else typeOfRenderer
        except: pass
        if typeOfRenderer == "continuationItemRenderer":
            continue
        eachFinal = RENDERER_PARSER[typeOfRenderer](each)
        result.append(eachFinal)
    if to_object: return SearchResult(result, nextCT)
    else: return result

class SearchResult(list):
    def __init__(self, data, continuation=None):
        super(SearchResult, self).__init__(data)
        self.continuation_token = continuation
        self.url = None

    def __add__(self, other):
        return SearchResult(list(self)+other)

    def __repr__(self):
        return f'<SearchResult {self.statistic}>'

    @property
    def statistic(self):
            try:
                result = dict(Counter(list(map(lambda i: i.__class__.__name__, self))))
                return result
            except: 
                return None

    @property
    def raw(self):
        return [i.raw for i in self]

class Mix:
    def __init__(self, data):
        self.id = data["playlistId"]
        self.title = data["title"]["simpleText"]
        self.video_count = "".join(i["text"] for i in data["videoCountShortText"]["runs"])
        self.videos = [PlaylistVideo(i["childVideoRenderer"]) for i in data["videos"]]
        self.thumbnail = data["thumbnail"]["thumbnails"]

    def __repr__(self):
        return f'<Mix id="{self.id}" title="{self.title}">'

    @property
    def raw(self):
        return {
            "type": "mix",
            "playlist_id": self.id,
            "title": self.title,
            "videos": [i.raw for i in self.videos],
            "video_count": self.video_count,
            "thumbnails": self.thumbnail
        }

class Shelf:
    def __init__(self, data):
        try: self.title = data["title"]["simpleText"]
        except: self.title = None
        self.videos = cleanupData(next(search_dict(data, "items")))

    def __repr__(self):
        return f'<Shelf title="{self.title}">'

    @property
    def raw(self):
        return {
            "type": "shelf",
            "title": self.title,
            "videos": [i.raw for i in self.videos]
        }

class LiveStream:
    def __init__(self, data):
            self.id = data["videoId"]
            self.title = "".join(i["text"] for i in data["title"]["runs"])
            self.description = "".join(i["text"] for i in data["descriptionSnippet"]["runs"]) if "descriptionSnippet" in data else None
            self.watching_count = int(data["viewCountText"]["runs"][0]["text"].replace(",", "")) if "viewCountText" in data else None
            self.author = Channel(
                name = data["ownerText"]["runs"][0]["text"],
                url = data["ownerText"]["runs"][0]["navigationEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"],
                channel_id = data["ownerText"]["runs"][0]["navigationEndpoint"]["browseEndpoint"]["browseId"],
            builtin_called = True
            )
            self.thumbnail = get_thumbnail(data["videoId"])

    def __repr__(self):
        return f'<LiveStream id="{self.id}" title="{self.title}">'

    @property
    def raw(self):
        return {
            "type": "live_stream",
            "videoId": self.id,
            "title": self.title,
            "description": self.description,
            "watching_count": self.watching_count,
            "author": self.author.raw,
            "thumbnails": self.thumbnail
        }

class HorizontalCardList:
    def __init__(self, data):
        title = next(search_dict(data["header"], "title"))
        if "simpleText" in title: title = title["simpleText"]
        elif "runs" in title: title = "".join(i["text"] for i in title["runs"])
        else: title = None
        self.title = title
        self.cards = cleanupData(data["cards"])

    def __repr__(self):
        return f'<HorizontalCardList title="{self.title}">'

    @property
    def raw(self):
        return {
            'title': self.title,
            'cards': self.cards
        }

class SearchRefinementCard:
    def __init__(self, data):
        self.query = "".join(i["text"] for i in data["query"]["runs"])
        self.url = next(search_dict(data, "url"))
        self.thumbnail = data["thumbnail"]["thumbnails"]

    def __repr__(self):
        return f'<SearchRefinementCard query="{self.query}">'

    @property
    def raw(self):
        return {
            "type": "search_refinement_card",
            "query": self.query,
            "url": self.url,
            "thumbnails": self.thumbnail
        }

class RichItem:
    def __init__(self, raw):
        pass

class BackgroundPromo:
    def __init__(self, data):
        self.title = "".join(i["text"] for i in data["title"]["runs"])
        self.content = "".join(i["text"] for i in data["bodyText"]["runs"])
    
    def __repr__(self):
        return f'<BackgroundPromo title="{self.title}">'

    @property
    def raw(self):
        return {
            "type": 'background_promo',
            "title": self.title,
            "content": self.content
        }

class Message:
    def __init__(self, data):
        self.text = "".join(i["text"] for i in data["text"]["runs"])
    
    def __repr__(self):
        return f'<Message text="{self.text}">'

    @property
    def raw(self):
        return {
            "type": "message",
            "text": self.text
        }

class SiteLinks:
    def __init__(self, title, url):
        self.title = title
        self.url = url

    def __repr__(self):
        return f'<SiteLinks title="{self.title}">'

    @property
    def redirected_url(self):
        return (requests.get(self.url).url)

class Advertisement:
    def __init__(self, data):
        self.title = data["title"]["simpleText"]
        self.description = data["descriptionText"]["simpleText"]
        self.website = "".join(i["text"] for i in data["websiteText"]["runs"])
        self.sitelinks = [SiteLinks("".join(i["text"] for i in i["title"]["runs"]), next(search_dict(i, 'url'))) for i in data["sitelinks"]] if 'sitelinks' in data else None

    def __repr__(self):
        return f'<Admertisement title="{self.title}">'

    @property
    def raw(self):
        return {
            "type": 'advertisement',
            "title": self.title,
            "description": self.description,
            "website": self.website
        }

class Playlist:
    def __init__(self, data):
        self.id = data["playlistId"]
        self.title = data["title"]["simpleText"]
        self.video_count = int(data["videoCount"])
        self.videos = [PlaylistVideo(i["childVideoRenderer"]) for i in data["videos"]]

    def __repr__(self):
        return f'<Playlist id="{self.id}" title="{self.title}">'

    @property
    def raw(self):
        return {
            "type": "playlist",
            "id": self.id,
            "title": self.title,
            "video_count": self.video_count,
            "videos": [i.raw for i in self.videos]
        }

class PlaylistVideo:
    def __init__(self, data):
        self.id = data["videoId"]
        self.title = data["title"]["simpleText"]
        self.length = data["lengthText"]["simpleText"]

    def __repr__(self):
        return f'<PlaylistVideo id="{self.id}" title="{self.title}">'

    @property
    def raw(self):
        return {
            "type": "playlist_video",
            "videoId": self.id,
            "title": self.title,
            "length": self.length
        }

class CarouselAd:
    def __init__(self, raw):
        pass

class DidYouMean:
    def __init__(self, data):
        self.query = data["correctedQuery"]["runs"][0]["text"]
        self.url = BASE_URL+next(search_dict(data, "url"))

    def __repr__(self):
        return f'<DidYouMean query="{self.query}">'

    @property
    def raw(self):
        return {
            'query': self.query,
            'url': self.url
        }

class ShowingResultsFor:
    def __init__(self, data):
        self.query = data["correctedQuery"]["runs"][0]["text"]
        self.url = BASE_URL+next(search_dict(data, "url"))
        self.original_query = BASE_URL+next(search_dict(data["originalQueryEndpoint"], "url"))

    def __repr__(self):
        return f'<ShowingResultsFor query="{self.query}">'

    @property
    def raw(self):
        return {
            'query': self.query,
            'url': self.url
        }

RENDERER_PARSER = {
    "videoRenderer": lambda x: Video(
        x["videoId"], 
        title="".join(i["text"] for i in x["title"]["runs"]),
        author=Channel(
            name = x["ownerText"]["runs"][0]["text"],
            url = x["ownerText"]["runs"][0]["navigationEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"],
            channel_id = x["ownerText"]["runs"][0]["navigationEndpoint"]["browseEndpoint"]["browseId"],
            builtin_called = True
        ),
        builtin_called=True
    ),
    "radioRenderer": Mix,
    "shelfRenderer": Shelf,
    "liveStreamRenderer": lambda x: Video(
        x["videoId"], 
        title="".join(i["text"] for i in x["title"]["runs"]),
        author=Channel(
            name = x["ownerText"]["runs"][0]["text"],
            url = x["ownerText"]["runs"][0]["navigationEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"],
            channel_id = x["ownerText"]["runs"][0]["navigationEndpoint"]["browseEndpoint"]["browseId"],
            builtin_called = True
        ),
        builtin_called=True
    ),
    "channelRenderer": lambda x: Channel(
        channel_id = x["channelId"],
        url = x["navigationEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"],
        name = x["title"]["simpleText"],
        video_count = int(x["videoCountText"]["runs"][0]["text"].split()[0].replace(",", "")) if "videoCountText" in x else None,
        subscriber_count = (int(d) if (d:=x["subscriberCountText"]["simpleText"].split()[0]).isdigit() else d) if "subscriberCountText" in x else None,
        avatar = x["thumbnail"]["thumbnails"],
        builtin_called = True
    ),
    "playlistRenderer": Playlist,
    "horizontalCardListRenderer": HorizontalCardList,
    "searchRefinementCardRenderer": SearchRefinementCard,
    "richItemRenderer": lambda x: cleanupData([x["content"]])[0],
    "backgroundPromoRenderer": BackgroundPromo,
    "messageRenderer": Message,
    "promotedSparklesTextSearchRenderer": lambda x: Advertisement(x['content']),
    "playlistVideoListRenderer": None,
    "playlistVideoRenderer": None,
    "carouselAdRenderer": None,
    "showingResultsForRenderer": lambda x: None,
    "previewCardRenderer": None,
    "searchPyvRenderer": lambda x: cleanupData(x['ads'])[0],
    "promotedVideoRenderer": lambda x: Video(
        x["videoId"], 
        title = x["title"]["simpleText"],
        author = Channel(
            name = x["longBylineText"]["runs"][0]["text"],
            url = x["longBylineText"]["runs"][0]["navigationEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"],
            channel_id = x["longBylineText"]["runs"][0]["navigationEndpoint"]["browseEndpoint"]["browseId"],
            builtin_called = True
        ),
        builtin_called=True
    ),
    'didYouMeanRenderer': DidYouMean,
    'showingResultsForRenderer': ShowingResultsFor,
    'premiereRenderer': lambda x: None
}


def custom_property(f):
    def wrapper(self, *args, **kwargs):
        if self._is_builtin_called and f.__name__ in self._static_properties:
            return self._static_properties[f.__name__]
        if not self._has_generated:
            self.parse_data()
        return f(self, *args, **kwargs)
    return property(wrapper, doc=f.__doc__)



from youtube_scraping_api.constants import HEADERS
from youtube_scraping_api.utils import search_dict, reveal_redirect_url, get_initial_data
from youtube_scraping_api.decorators import custom_property
from youtube_scraping_api.urls import *

import requests
import re

class Channel:
    """Container of channel data"""
    def __init__(self, channel_id=None, username=None, builtin_called=False, **kwargs):
        self._session = requests.Session()
        self._session.headers = HEADERS
        self._has_generated = False

        self.id = channel_id
        self.username = username

        self._about_data = None
        self._metadata = None
        self._header_data = None
        self._is_builtin_called = builtin_called

        self._static_properties = kwargs

    def __repr__(self):
        return f'<Channel id="{self.id}" name="{self.name}">'

    def parse_data(self):
        """Fetch HTML source code and extract JSON data from it

        :return: Nothing, data have been set inside local variable
        :rtype: None
        """
        if not (self.id or self.username): return {}
        if self.id: url = (
            CHANNEL_ID_URL+self.id,
            CHANNEL_ID_URL+self.id+"/about"
        )
        elif self.username: url = (
            CHANNEL_USERNAME_URL+self.username,
            CHANNEL_USERNAME_URL+self.username+"/about"
        )
        response = [self._session.get(i).text for i in url]
        if "404 Not Found" in response[0]:
            self._debug("ERROR", "Channel not exist")
            return
        data, self._about_data = (get_initial_data(i) for i in response)
        self._metadata = data["metadata"]["channelMetadataRenderer"]
        self._header_data = data["header"]

        self._has_generated = True

    @custom_property
    def name(self):
        """Extract name of the channel

        :return: Channel name
        :rtype: str
        """
        return self._metadata["title"]

    @custom_property
    def description(self):
        """Extract description of the channel

        :return: Channel description
        :rtype: str
        """
        return self._metadata["description"]

    @custom_property
    def keywords(self):
        """Extract descriptive keywords which content creators can add to thier channel to help viewers find them

        :return: List of all keywords of the channel
        :rtype: list
        """
        if not 'keywords' in self._metadata: return None
        patterns = [
            r'(?:^|\"\s)([^\"].*?[^\"])(?:\s\"|$)',
            r'\".*?\"'
        ]
        raw_keywords = self._metadata["keywords"]
        keywords_list = sum([re.findall(pattern, raw_keywords) for pattern in patterns], [])
        final = [[eval(keyword)] if re.match(patterns[1], keyword) else keyword.split() for keyword in keywords_list]
        return sum(final, [])

    @custom_property
    def url(self):
        """Return url of the channel

        :return: Channel url
        :rtype: str
        """
        return self._metadata["channelUrl"]

    @custom_property
    def vanity_url(self):
        """Return a human readable custom url created by the channel owner if available

        :return: Custom channel url
        :rtype: str
        """
        return self._metadata["vanityChannelUrl"]

    @custom_property
    def facebook_profile_id(self):
        """Extract facebook profile id of channel owner if available

        :return: Facebook profile id of the channel owner
        :rtype: str or None
        """
        return self._metadata["facebookProfileId"] if "facebookProfileId" in self._metadata else None

    @custom_property
    def avatar(self):
        """Extract avatar thumbnail url of the channel

        :return: Avatar thumbnail url of the channel
        :rtype: str
        """
        return self._metadata["avatar"]["thumbnails"][0]

    @custom_property
    def subscriber_count(self):
        """Extract number of subscriber of the cannel

        :return: Number of subscriber of the channel
        :rtype: int or str or None
        """
        try: subscriber_count = next(search_dict(self._header_data, "subscriberCountText"))["simpleText"].split()[0]
        except: subscriber_count = "No"
        return int(subscriber_count) if subscriber_count.isdigit() else None if subscriber_count == "No" else subscriber_count

    @custom_property
    def banner(self):
        """Return a list of banner urls in different resolutions if available

        :return: A list of banner urls
        :rtype: list or None
        """
        try: return next(search_dict(self._header_data, "banner"))["thumbnails"]
        except: return None

    @custom_property
    def header_links(self):
        """Return a list of social media links that content creator put on their channel header if available

        :return: A list of social media links
        :rtype: list or None
        """
        try:
            raw_header_links = next(search_dict(self._header_data, "channelHeaderLinksRenderer")).values()
            header_links = [{
                "title": i["title"]["simpleText"],
                "icon": i["icon"]["thumbnails"][0]["url"],
                "url": reveal_redirect_url(i["navigationEndpoint"]["urlEndpoint"]["url"])
            } for i in sum(raw_header_links, [])]
            return header_links
        except:
            return None

    @custom_property
    def is_verified(self):
        """Check if the channel is verified by YouTube

        :return: A boolean that states whether the channel is verified
        :rtype: bool
        """
        try:
            if next(search_dict(self._header_data, "badges"))[0]["metadataBadgeRenderer"]["tooltip"] == "Verified":
                return True
            return False
        except: return False

    @custom_property
    def raw(self):
        """Returns all available channel metadata

        :return: A dictionary with metadata values
        :rtype: dict
        """
        return {
            'name': self.name,
            'description': self.description,
            'keywords': self.keywords,
            'url': self.url,
            'vanity_url': self.vanity_url,
            'facebook_profile_id': self.facebook_profile_id,
            'avatar': self.avatar,
            'subscriber_count': self.subscriber_count,
            'header_links': self.header_links,
            'is_verified': self.is_verified
        }



from youtube_scraping_api.parser.search_result import SearchResult, cleanupData as cleanupVideoData
from youtube_scraping_api.parser.playlist import Playlist, cleanupData as cleanupPlaylistData
from youtube_scraping_api.constants import PAYLOAD, HEADERS
from youtube_scraping_api.parser.channel import Channel
from youtube_scraping_api.parser.video import Video
from youtube_scraping_api.decorators import *
from youtube_scraping_api.filter import *
from youtube_scraping_api.utils import *
from youtube_scraping_api.urls import *

import itertools
import requests

class YoutubeAPI:
    """Core developer interface for Youtube Scraping API
    """
    _data = PAYLOAD

    def __init__(self):
        self._session = requests.Session()
        self._session.headers = HEADERS
        try: raw = self._session.get("https://www.youtube.com").text
        except: raise RuntimeError("Please check your internet connection")

        if len(raw) < 10000: 
            self._session.proxies = get_proxy()
        
        self.API_TOKEN = find_snippet(raw, "innertubeApiKey", ",", (3, 1))

    def search(self, query=None, continuation_token=None, raw=False, filter=None):
        """Parse YouTube search results of specific query or continuation token

        :param query: A query string to search on YouTube
        :type query: str, optional
        :param continuation_token: A token generated by YouTube to fetch more search results
        :type continuation_token: str, optional
        :param raw: Whether to return search results in raw format. Default set to False
        :type raw: bool, optional
        :param filter: Filter for search results
        :type filter: SearchFilter, optinal
        :return: Search results
        :rtype: SearchResult
        """
        if not (query or continuation_token): return None

        if query:
            base_url = SEARCH_BASE_URL+"+".join(query.split())
            if filter and isinstance(filter, SearchFilter): final_url = get_filtered_url(self._session, base_url, filter)
            else: final_url = base_url
            html = self._session.get(final_url).text
            response = get_initial_data(html)
        
        elif continuation_token:
            self._data["continuation"] = continuation_token
            response = self._session.post(SEARCH_CONTINUATION_URL+self.API_TOKEN, json=self._data).json()

        nextCT = parse_continuation_token(response)
        
        if query:
            data = [next(search_dict(i, "contents")) for i in search_dict(response,"itemSectionRenderer")]
            result = SearchResult(itertools.chain(*[cleanupVideoData(i) for i in data]), nextCT)
            result.url = final_url
        
        if continuation_token:
            try: data = next(search_dict(response, "contents"))
            except: data = next(search_dict(response, "continuationItems"))
            result = cleanupVideoData(data, nextCT, to_object=True)
        
        if not raw: return result
        else: return result.raw

    def playlist(self, playlist_id=None, continuation_token=None):
        """Parse playlist metadata and videos

        :param playlistId:
            ID of playlist
        :type playlistId: str, optional
        :param continuation_token:
            A token generated by YouTube to fetch more playlist videos
        :type continuation_token: str, optional
        :return: Playlist Object
        :rtype: Playlist
        """
        if not (playlist_id or continuation_token): return None

        if playlist_id:
            html = self._session.get(PLAYLIST_BASE_URL+playlist_id).text
            response = get_initial_data(html)

        elif continuation_token:
            if not continuation_token: return {}, None
            self._data["continuation"] = continuation_token
            response = self._session.post(PLAYLIST_CONTINUTION_URL+self.API_TOKEN, json=self._data).json()

        nextCT = parse_continuation_token(response)

        if playlist_id: 
            result = Playlist(response)
        
        elif continuation_token: 
            data = next(search_dict(response, "continuationItems"))
            result = cleanupPlaylistData(data)

        while nextCT:
            response, nextCT = self.playlist(continuation_token = nextCT)
            result += response

        if playlist_id: return result
        elif continuation_token: return result, nextCT

    def channel(self, channel_id=None, username=None):
        """Parse channel metadata and all it's videos

        :param channel_id:
            ID of channel
        :type channel_id: str, optional
        :param username:
            Username of channel owner
        :type username: str, optional
        :return: Channel Object
        :rtype: Channel
        """
        result = Channel(channel_id=channel_id, username=username)
        return result

    def video(self, videoId: str) -> Video:
        """Parse video metadata, captions, download link, etc.

        :param videoId:
            ID of Youtube video
        :type videoId: str
        :return: Video object
        :rtype: Video
        """
        return Video(videoId)

    def query_suggestions(self, query=None, language='en', country='gb'):
        """Return a list of query suggestions for given query string

        :param query: A string of query, defaults set to None
        :type query: str, optional
        :param language: Language of results, defaults set to 'en'
        :type language: str, optional
        :param country: Country code for more accurate suggestions, defaults set to 'gb'
        :type country: str, optional
        :return: A list of query suggestions
        :rtype: list
        """
        parameter = dict(client='youtube', hl=language, gl=country, q=query)
        raw_data = self._session.get(QUERY_SUGGESTIONS_URL, params=parameter).text
        data = eval(raw_data[raw_data.find('('):])[1]
        result = [i[0] for i in data]
        return result
